#!/usr/bin/env python

"""
Module processes Json-Files that were generated with e.g. Hydration.hydrate_file containing one tweet per line. 
The information gets cleaned and processed such that it can be read and transformed via the pandas-library.
"""

import json
import logging
import re
import os
from datetime import datetime
import pytz


def process_tweets(input_path, output_path):
    """
    Takes input-path of file in Json-Format containing on each line information for exactly one tweet and transforms it
    to Csv-File containing only the relevant data. It contains the following headers:
    
    lang - Language of tweet
    
    created_at - Date of creation of tweet
    
    rt_count - Number of times retweeted
    
    fav_count - Number of times this tweet has been liked
    
    t_id - tweet id
    
    statuses_count - Number of tweets tweeted by the user 
    
    profile_fav_count - Number of times the users profile has been liked
    
    followers_count - Amount of followers
    
    friends_count - Amount of friends
    
    listed_count - Number of public lists this user is member of
    
    screen_name - Screen name of user
    
    u_id - User identification number
    
    is_geo - Does tweet contain geo-information? (1: true, 0: false)
    
    geo_lat - Longitude of coordinate

    geo_long - Latitude of coordinate
    
    is_reply - Is tweet a reply? (1: true, 0: false)
    
    in_reply_to_screen_name - Screen name of user who's tweet this tweet was replying to
    
    in_reply_to_status_id - Identification number of tweet this tweet was replying to
    
    in_reply_to_user_id_str - Identifation number string of tweet this tweet was replying to
    
    RT - Is tweet a retweet? (1: true, 0: false)
    
    r_id - Identification number of retweeted (original) tweet 
    
    r_screen_name - Screen name of user whos tweet has been retweeted
    
    r_uid - User identification number of user whos tweet has been retweeted
    
    QT - Is tweet a quote of another tweet? (1: true, 0: false)
    
    q_id - Identification number of quoted (original) tweet
    
    q_screen_name - Screen name of user whos tweet has been quoted
    
    q_uid - User identification number of user whos tweet has been quoted
    
    hashtags - Used hashtags
    
    usermentions - Mentioned users
    
    urls - Contained Urls
    
    media - Does tweet contain Urls leading to media files? (1: true, 0: false)
    
    mediaurls - Contained Urls leading to media-files
    
    text - Tweet text
    
    :param str input_path: Path to Json-File containing tweet information
    
    :param str output_path: Path to destination of Csv-File
    
    """

    # Define Statistic Values
    lineNum = 1
    deleted_tweet_count = 0
    wrong_lang = 0
    lastid = "None"

    # Open input and output-file
    i = open(input_path)
    o = open(output_path, 'w')

    # Set up logger
    dir_name, _ = os.path.splitext(output_path)
    logger = logging.getLogger(dir_name)
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler("{0}.log".format(dir_name), mode='w')
    fh.setLevel(logging.INFO)
    fh_format = logging.Formatter('%(asctime)s, %(message)s')
    fh.setFormatter(fh_format)
    logger.addHandler(fh)

    # Write header of log-file
    fh.setFormatter(logging.Formatter('%(message)s'))
    logger.info("time,is_deleted,tweet_id,deleted_tweet_user_id")
    fh.setFormatter(logging.Formatter('%(asctime)s,%(message)s'))
    header = "lang,created_at,rt_count,fav_count,t_id,statuses_count,profile_fav_count"
    header += ",followers_count,friends_count,listed_count,screen_name,u_id"
    header += ",is_geo,geo_lat,geo_long"
    header += ",is_reply,in_reply_to_screen_name,in_reply_to_status_id,in_reply_to_user_id_str"
    header += ",RT,r_id,r_screen_name,r_uid"
    header += ",QT,q_id,q_screen_name,q_uid"
    header += ",hashtags,usermentions,urls"
    header += ",media,mediaurls"
    header += ",text"
    o.write(header + '\n')
    o.flush()

    # Iterate through input-file and extract the wanted information, print it to output-file in csv-wise manner
    for line in i.readlines():
        try:
            tweet = json.loads(line)
        except json.decoder.JSONDecodeError as e:
            logger.info("1,Json Error on tweet,Position:" + str(e.pos)+', line:'+str(e.lineno))
            deleted_tweet_count = deleted_tweet_count + 1
            continue
        if 'delete' in tweet:
            logger.info("1," + tweet['delete']['status']['id_str'] + "," + tweet['delete']['status']['user_id_str'])
            deleted_tweet_count = deleted_tweet_count + 1
        else:
            if tweet['lang'] != "en":
                wrong_lang = wrong_lang + 1
            else:
                pstr = tweet['lang']
                pstr += "," + str(datetime.strptime(tweet['created_at'], '%a %b %d %H:%M:%S +0000 %Y').replace(tzinfo=pytz.UTC))
                pstr += "," + str(tweet['retweet_count'])
                pstr += "," + str(tweet['favorite_count'])
                pstr += "," + tweet['id_str']
                pstr += "," + str(tweet['user']['statuses_count'])
                pstr += "," + str(tweet['user']['favourites_count'])
                pstr += "," + str(tweet['user']['followers_count'])
                pstr += "," + str(tweet['user']['friends_count'])
                pstr += "," + str(tweet['user']['listed_count'])
                pstr += "," + tweet['user']['screen_name']
                pstr += "," + tweet['user']['id_str']

                if tweet['coordinates'] is not None and tweet['coordinates']['type'] == 'Point':
                    pstr += ',1,' + str(tweet['coordinates']['coordinates'][1]) + "," + str(
                        tweet['coordinates']['coordinates'][0])
                elif tweet['coordinates'] is None and tweet['place'] is not None:
                    if tweet['place']['bounding_box'] is not None:
                        coordinates = tweet['place']['bounding_box']['coordinates'][0]
                        longitude = (int(coordinates[0][0]) + int(coordinates[1][0]) + int(coordinates[2][0]) + int(
                            coordinates[3][0])) / 4
                        latitude = (int(coordinates[0][1]) + int(coordinates[1][1]) + int(coordinates[2][1]) + int(
                            coordinates[3][1])) / 4
                        pstr += ',1,' + str(latitude) + "," + str(longitude)
                elif tweet['coordinates'] is None and tweet['place'] is None:
                    pstr += ',0,,'

                if tweet['in_reply_to_status_id_str'] is None:
                    pstr += ',0,,,'
                else:
                    pstr += ",1" + "," + str(tweet['in_reply_to_screen_name']) + "," + str(
                        tweet['in_reply_to_status_id_str']) + "," + str(tweet['in_reply_to_user_id_str'])

                if 'retweeted_status' in tweet:
                    pstr = pstr + ",1," + tweet['retweeted_status']['id_str']
                    pstr += "," + tweet['retweeted_status']['user']['screen_name']
                    pstr += "," + tweet['retweeted_status']['user']['id_str']
                else:
                    pstr = pstr + ",0,,,"
                if 'quoted_status' in tweet:
                    pstr = pstr + ",1," + tweet['quoted_status']['id_str']
                    pstr += "," + tweet['quoted_status']['user']['screen_name']
                    pstr += "," + tweet['quoted_status']['user']['id_str']
                else:
                    pstr = pstr + ",0,,,"

                hashtags = [a['text'] for a in tweet['entities']['hashtags']]
                usermentions = [a['id_str'] for a in tweet['entities']['user_mentions']]
                urls = [a['expanded_url'] for a in tweet['entities']['urls'] if a['expanded_url'] != 'null']
                urls = [re.sub("\b[']\b'", string=a, repl='') for a in urls]
                urls = str(urls).replace('"', "'")
                pstr = pstr + ',"' + str(hashtags) + '","' + str(usermentions) + '","' + str(urls) + '"'
                if 'media' in tweet['entities']:
                    mediaurls = [media['media_url'] for media in tweet['entities']['media']]
                    mediaurls = [re.sub("\b[']\b'", string=a, repl='') for a in mediaurls]
                    mediaurls = str(mediaurls).replace('"', "'")
                    pstr += ',1,"' + str(mediaurls) + '"'
                else:
                    pstr += ',0,'
                tweet_text = tweet['text'].replace('\n', ' ').replace('\r', '')
                tweet_text = tweet_text.replace('"', "'")
                pstr += ',"' + tweet_text + '"'
                pstr.encode('utf8')
                lastid = tweet['id_str']
                o.write(pstr + '\n')
                o.flush()
                lineNum += 1

    # Write Summary to logfile
    fh.setFormatter(logging.Formatter('%(message)s'))
    logger.info("\n")
    logger.info("----------------------------------------------------------------")
    logger.info("\n")
    logger.info("Summary")
    logger.info("Total Tweets Processed (Input): " + str(lineNum - 1))
    logger.info("Total Tweets Processed (Output): " + str((lineNum - 1) - (deleted_tweet_count + wrong_lang)))
    logger.info("Total Deleted Tweets: " + str(deleted_tweet_count))
    logger.info("Wrong Language Tweets: " + str(wrong_lang))
    logger.info("Last Hydrated Tweet ID: " + str(lastid))
    logging.shutdown()




"""
A Factory to handle and cluster all networkx-based graph activities. Contains methods for creating, pruning, 
plotting and information-gathering.
"""

import itertools
import operator
from collections import OrderedDict
from itertools import islice
from multiprocessing import Pool
import community
import matplotlib

matplotlib.use('Agg')
import matplotlib.cm as cmx
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import networkx as nx
import pandas as pd
from networkx.drawing.nx_agraph import graphviz_layout
from communitweet.ForceAtlas2 import *
from communitweet.Summarize import *

class GraphFactory:
    """
    GraphFactory-Class contains all relevant class-variables and methods. Represents a nx.Graph()-object
    """
    def __init__(self, name_of_graph):
        """
        Initializes graph and gives name to graph
        
        :param str name_of_graph: Name of graph needs to be specified here
        
        """
        print('Start Creating Graph "' + name_of_graph + '"')
        self.graph = nx.Graph()
        self.name_of_graph = name_of_graph
        self.pruning_information = self._PruningInfo()
        self.community_partitions = dict()
        self.dataframe = None

    def add_node(self, node, label):
        """
        Method to add node and its label to the graph-object. Handles increasing of 'weight' attribute of each node
        automatically each time a node gets added multiple times
        
        :param int node: Node in the form of a twitter user-id or a hashtag
        
        :param str label: Label of node, either screen_name of a user or the name of the hashtag
        
        """
        node = int(node)
        label = str(label)
        if not self.graph.has_node(node):
            self.graph.add_node(node, attr_dict={'weight': 1, 'label': label})
        else:
            self.graph.node[node]['weight'] += 1
            self.graph.node[node]['label'] = label

    def add_edge(self, node_1, label_1, node_2, label_2):
        """
        Method to add edge to the graph-object. Handles increasing of 'weight' attribute of each edge and the related 
        nodes automatically each time an edge gets added multiple times.
        
        :param int node_1: First node or edge start node
        
        :param str label_1: Label of first node
        
        :param int node_2: Second node or edge end node
        
        :param str label_2: Label of second node
        
        """
        node_1 = int(node_1)
        node_2 = int(node_2)
        label_1 = str(label_1)
        label_2 = str(label_2)

        if not self.graph.has_edge(node_1, node_2):
            if not self.graph.has_node(node_1):
                self.graph.add_node(node_1, attr_dict={'weight': 1, 'label': label_1})
            if not self.graph.has_node(node_2):
                self.graph.add_node(node_2, attr_dict={'weight': 1, 'label': label_2})
            self.graph.add_edge(node_1, node_2, attr_dict={'weight': 1})
        else:
            self.graph[node_1][node_2]['weight'] += 1

    def write_graphml(self, output_path):
        """
        Writes the graph-object to file in the form of a Graphml-File
        
        :param str output_path: Path to destination of Graphml-File
        
        """
        print("Writing Graph to " + str(output_path))
        nx.write_graphml(self.graph, output_path)

    def load_graph(self, input_path):
        """
        Loads a graph-object from a Graphml-File
        
        :param str input_path: path to Graphml-File
        
        """
        print("Loading Graph from " + str(input_path))
        with open(input_path) as inp:
            self.graph = nx.read_graphml(inp)

    def print_info(self):
        """
        Prints a short overview of the graph including the amount of nodes, edges, the average degree and the average 
        weight
        """
        print('--------------------------------------')
        print('Information on ' + '"' + self.name_of_graph + '"' + '-Graph')
        print('Number of nodes: ' + str(self.graph.number_of_nodes()))
        print('Number of edges: ' + str(self.graph.number_of_edges()))
        print('Average degree: ' + str(round(float(sum(self.graph.degree().values())) / float(self.graph.number_of_nodes()),2)))

        avg_sum = 0
        for _, data in self.graph.nodes(data=True):
            avg_sum += data['weight']

        print('Average weight: ' + str(round(float(avg_sum) / float(self.graph.number_of_nodes()),2)))
        print('--------------------------------------')

    def get_threshold(self, total_percentage_of_nodes_remaining, blur, name_of_measure):
        """
        Calculates the threshold for a given name_of_measure such that, when specified measure is pruned according to 
        that threshold, the percentage of nodes remaining in the graph is equal to the parameter total_percentage_of_nodes_remaining.
        Calculated threshold is only right, when pruning option remove_isolated is applied separately.
        
        :param int total_percentage_of_nodes_remaining: Percentage of nodes to keep in the graph after pruning
        
        :param int blur: How many percentage-points can the threshold be wrong
        
        :param str name_of_measure: On what node measure should the algorithm be applied. So far only 'weight_thresh' or 'degree_thresh' are possible
        
        :return int: 
        """
        total_node_count = self.graph.number_of_nodes()
        if total_node_count == 0:
            return 0
        if name_of_measure == 'weight_thresh':
            temp_threshold = 2
            continue_eval = True
            nodes_to_remove = []

            error_count_minus = 0
            error_count_plus = 0
            while continue_eval:
                for node, data in self.graph.nodes(data=True):
                    if data['weight'] < temp_threshold:
                        nodes_to_remove.append(node)
                temp_remaining = 100 - len(nodes_to_remove) / total_node_count * 100
                if total_percentage_of_nodes_remaining + blur > temp_remaining > total_percentage_of_nodes_remaining - blur:
                    continue_eval = False
                elif total_percentage_of_nodes_remaining - blur < temp_remaining:
                    temp_threshold = temp_threshold + 1
                    error_count_plus += 1
                elif temp_remaining < total_percentage_of_nodes_remaining - blur:
                    temp_threshold = temp_threshold - 1
                    error_count_minus += 1
                if error_count_plus > 2 and error_count_minus > 2:
                    error_count_plus = 0
                    error_count_minus = 0
                    continue_eval = False
                nodes_to_remove = []
            return temp_threshold
        if name_of_measure == 'degree_thresh':
            temp_threshold = 2
            continue_eval = True
            nodes_to_remove = []

            error_count_minus = 0
            error_count_plus = 0
            while continue_eval:
                for node, data in self.graph.nodes(data=True):
                    if 'Degree' not in data:
                        print('Graph has no attribute "degree" - please use "add_stats"-function')
                        exit(1)
                    if data['Degree'] < temp_threshold:
                        nodes_to_remove.append(node)
                temp_remaining = 100 - len(nodes_to_remove) / total_node_count * 100
                if total_percentage_of_nodes_remaining + blur > temp_remaining > total_percentage_of_nodes_remaining - blur:
                    continue_eval = False
                elif total_percentage_of_nodes_remaining - blur < temp_remaining:
                    temp_threshold = temp_threshold + 1
                    error_count_plus += 1
                elif temp_remaining < total_percentage_of_nodes_remaining - blur:
                    temp_threshold = temp_threshold - 1
                    error_count_minus += 1
                if error_count_plus > 2 and error_count_minus > 2:
                    error_count_plus = 0
                    error_count_minus = 0
                    continue_eval = False
                nodes_to_remove = []
            return temp_threshold

    def prune_graph(self, edge_weight, degree_thresh, weight_thresh, remove_isolates):
        """
        Prunes the graph-object on the specified measurements.
        
        :param int edge_weight: minimum weight an edge can have to remain in the graph
        
        :param int degree_thresh: minimum degree an node can have to remain in the graph
        
        :param int weight_thresh: minimum weight an node can have to remain in the graph
        
        :param bool remove_isolates: should isolated nodes be removed?
        
        """
        print("Pruning Graph")
        print(".......... Removing self refering edges")

        if edge_weight != 0:
            self.pruning_information.edge_weight = edge_weight
        if degree_thresh != 0:
            self.pruning_information.node_degree = degree_thresh
        if weight_thresh != 0:
            self.pruning_information.node_weight = weight_thresh
        if remove_isolates:
            self.pruning_information.isolated = remove_isolates

        edges_to_remove = []
        for edge in self.graph.edges(data=True):
            if edge[0] == edge[1]:
                edges_to_remove.append(edge)
            if edge[2]['weight'] < edge_weight:
                edges_to_remove.append(edge)

        self.graph.remove_edges_from(edges_to_remove)
        print(".......... Removing nodes according to degree and weight")

        nodes_to_remove = []
        for node, data in self.graph.nodes(data=True):
            if data['weight'] < weight_thresh:
                nodes_to_remove.append(node)
            elif data['Degree'] < degree_thresh:
                nodes_to_remove.append(node)

        self.graph.remove_nodes_from(nodes_to_remove)

        if remove_isolates:
            print(".......... Removing isolated nodes")
            self.graph.remove_nodes_from(nx.isolates(self.graph))

        # components = sorted(nx.connected_component_subgraphs(self.graph),key=len,reverse=True)
        # self.graph = self.graph.subgraph(components[0:10],inplace=True)

        print("Pruning Done")

    class _PruningInfo():
        """
        Class to pack all the Information on the applied Pruning that has been done on current graph-object.
        """
        def __init__(self):
            """
            Sets all pruning-info to 0.
            """
            self.node_weight = 0
            self.node_degree = 0
            self.edge_weight = 0
            self.isolated = False

        def to_string(self):
            """
            Returns all the info contained in _PruningInfo in one string.
            """
            output = 'Minimum Edge Weight = ' + str(self.edge_weight) + '\n'
            output += 'Minimum Degree of Nodes = ' + str(self.node_degree) + '\n'
            output += 'Minimum Weight of Nodes = ' + str(self.node_weight) + '\n'
            output += 'Isolated Nodes were removed = ' + str(self.isolated)
            return output

    def add_stats(self, degree, closeness, betweenness):
        """
        Calculates measures and statistics such as degree, closeness and betweenness to the graph-object and to its
        nodes.
        
        :param bool degree: If set to true, the degree of each node is added as an attribute
        
        :param bool closeness: If set to true, the closeness of each node is added as an attribute
        
        :param bool betweenness: If set to true, the betweeness of each node is added as an attribute. ATTENTION: Takes a long time
        
        """
        print("Calculating Stats")
        if degree:
            print(".......... Degree")
            deg = self.graph.degree()
            nx.set_node_attributes(self.graph, 'Degree', deg)
        if closeness:
            print(".......... Closeness")
            close = nx.closeness_centrality(self.graph)
            nx.set_node_attributes(self.graph, 'Closeness_centrality', close)
        if betweenness:
            print(".......... Betweeness")
            betwe = calculate_betweenness(self.graph)
            nx.set_node_attributes(self.graph, 'Betweenness_centrality', betwe)
        print("Calculating Degree")

    def _get_communities(self):
        """
        Helper method to calculate communities with the help of the 'community'-package
        """
        self.community_partitions = community.best_partition(self.graph)

    def write_out_community_user_info(self, path_to_file):
        """
        Calculates communities - if not already done - and sorts each community by weight as well as all communities 
        by size, analyzes the graph and prints all information to TXT-file.
        The generated file contains information on the whole graph - including the top 10 hashtags, locations, urls and media-urls
        of all users of the graph as welll as of the users of each community.
        
        :param str path_to_file: Path to destination for Information-File.
        """
        print('Evaluating communites')
        if not self.community_partitions:
            self._get_communities()

        all_communities_sorted = dict()
        communities = set([comm for comm in self.community_partitions.values()])
        for comm in communities:
            temp_user_list = []
            for k, v in self.community_partitions.items():
                if v == comm:
                    temp_user_list.append(k)
            all_communities_sorted[comm] = temp_user_list

        # sort communities according to size
        print('Sorting communites')
        all_communities_sorted = OrderedDict(
            sorted(all_communities_sorted.items(), key=lambda x: len(x[1]), reverse=True))

        # get top 10 of users according to weight
        list_of_nodes_with_weights = dict(
            [(node, (data['weight'], data['label'])) for (node, data) in self.graph.nodes(data=True)])

        all_communities_sorted_with_weight = OrderedDict()
        for comm, users_of_communities in all_communities_sorted.items():
            temp_list = []
            for user in users_of_communities:
                temp_list.append((user, list_of_nodes_with_weights[user]))
            temp_list.sort(key=operator.itemgetter(1), reverse=True)
            all_communities_sorted_with_weight[comm] = temp_list

        total_info_dict = OrderedDict()
        for comm, users_of_communities_with_weight in all_communities_sorted_with_weight.items():
            temp_list = [(a[0], a[1][1], a[1][0]) for a in users_of_communities_with_weight]
            temp_list = temp_list[0:10]
            total_info_dict[comm] = temp_list

        # create separate dataframes for each community and calculate top results for hashtags, links and media
        if self.dataframe is not None:
            for comm, community_members in total_info_dict.items():
                list_of_ids = []
                for member in community_members:
                    list_of_ids.append(member[0])
                comm_dataframe = self.dataframe[self.dataframe.u_id.isin(list_of_ids)]
                top_10_hashtags = get_top_n_hashtags(10, comm_dataframe, give_freq=True)
                top_10_media = get_top_n_media(10, comm_dataframe, give_freq=True)
                top_10_urls = get_top_n_urls(10, comm_dataframe, give_freq=True)
                top_10_locations = get_top_n_locations(10, comm_dataframe, give_freq=True)
                total_info_dict[comm] = {'top_users': total_info_dict[comm], 'top_10_hashtags': top_10_hashtags,
                                         'top_10_media': top_10_media, 'top_10_urls': top_10_urls,
                                         'state_fips': top_10_locations}

        # create string containing file-information
        print('Creating Info-File')
        outstring = 'Information on Dataset:' + '\n\n'

        outstring += '   Name: ' + self.name_of_graph + '\n'
        outstring += '   Number of nodes: ' + str(self.graph.number_of_nodes())+ '\n'
        outstring += '   Number of edges: ' + str(self.graph.number_of_edges())+ '\n'
        if self.graph.number_of_nodes() > 0:
            outstring += '   Average degree: ' + str(float(sum(self.graph.degree().values())) / float(self.graph.number_of_nodes()))+ '\n'
            avg_sum = 0
            for _, data in self.graph.nodes(data=True):
                avg_sum += data['weight']

            outstring += '   Average weight: ' + str(float(avg_sum) / float(self.graph.number_of_nodes()))+ '\n\n'

        if self.dataframe is not None:
            # TODO: So far: Nur of times tweeted - To add: Nr of times retweeted
            analyze_info = analyze_dataframe(self.dataframe, True)
            outstring += '   Top 10 users:' + '\n'
            for (user, name), weight in analyze_info['users']:
                outstring += '      ' + str(name) + '(' + str(user) + ') - Count: ' + str(weight) + '\n'
            outstring += '\n'
            outstring += '   Top 10 hashtags \n'
            for hashtag, weight in analyze_info['hashtags']:
                outstring += '      ' + str(hashtag) + ' - Count: ' + str(weight) + '\n'
            outstring += '\n'

            outstring += '   Top 10 Urls \n'
            for url, weight in analyze_info['urls']:
                outstring += '      ' + str(url) + ' - Count: ' + str(weight) + '\n'
            outstring += '\n'

            outstring += '   Top 10 media (Percentage of total who have media: ' + str(
                analyze_info['media_perc']) + ')\n'
            for media, weight in analyze_info['media']:
                outstring += '      ' + str(media) + ' - Count: ' + str(weight) + '\n'
            outstring += '\n'

            outstring += '   Top 10 locations FIPS (Percentage of total: ' + str(analyze_info['stats_perc']) + ')\n'
            for (_, fips), weight in analyze_info['state_fips']:
                outstring += '      ' + str(fips) + ' - Count: ' + str(weight) + '\n'
            outstring += '\n'

            outstring += '   Percentage of retweets: ' + str(analyze_info['rt_perc']) + '\n'
            outstring += '\n'
            outstring += '   Percentage of quotes: ' + str(analyze_info['qt_perc']) + '\n'
            outstring += '\n'
            outstring += '   Percentage of answers: ' + str(analyze_info['reply_perc']) + '\n'
            outstring += '\n'
            outstring += '   Percentage of demographic statistics: ' + str(analyze_info['stats_perc']) + '\n'
            outstring += '\n'
            outstring += '   Percentage of geographic information: ' + str(analyze_info['geo_perc']) + '\n'
            outstring += '\n\n'
            outstring += 'Communities (Ordered after size): \n'
            outstring += 'Number of communities: ' + str(len(communities)) + '\n'

        for comm, info_dict in total_info_dict.items():

            outstring += '\n'
            outstring += 'Community ' + str(comm) + ':' + '\n'

            if 'top_users' in info_dict:
                if info_dict['top_users']:
                    outstring += '   Top 10 users:' + '\n'
                    for (screen_id, label, weight) in info_dict['top_users']:
                        outstring += '      ' + label + '(' + str(screen_id) + ')' + ' - Count: ' + str(weight) + '\n'
                    outstring += '\n'
            if 'top_10_hashtags' in info_dict:
                if info_dict['top_10_hashtags']:
                    outstring += '   Top 10 hashtags' + '\n'
                    for hashtag, weight in info_dict['top_10_hashtags']:
                        outstring += '      ' + str(hashtag) + ' - Count: ' + str(weight) + '\n'
                    outstring += '\n'
            if 'top_10_media' in info_dict:
                if info_dict['top_10_media']:
                    outstring += '   Top 10 media' + '\n'
                    for media, weight in info_dict['top_10_media']:
                        outstring += '      ' + str(media) + ' - Count: ' + str(weight) + '\n'
                    outstring += '\n'
            if 'top_10_urls' in info_dict:
                if info_dict['top_10_urls']:
                    outstring += '   Top 10 Urls' + '\n'
                    for url, weight in info_dict['top_10_urls']:
                        outstring += '      ' + str(url) + ' - Count: ' + str(weight) + '\n'
                    outstring += '\n'

            if 'state_fips' in info_dict:
                if info_dict['state_fips']:
                    outstring += '   Top 10 States' + '\n'
                    for (fips,name), weight in info_dict['state_fips']:
                        outstring += '      ' + str(name) + ' - Count: ' + str(weight) + '\n'
                    outstring += '\n'

        with open(path_to_file, 'w') as file_out:
            file_out.writelines(outstring)

    def create_graphs_from_dataframe(self, on_key, path_to_csv):
        """
        Method to quickly generate graphs on QT, RT or hashtags. If the method write_out_community_user_info should be 
        applied, the graph needs to be generated with this method or else there will be no underlying dataframe 
        to gather information from.

        
        :param str on_key: On which key should the graph be generated. Possible are retweeted('RT'), quoted('QT') or hashtags('hashtags')
        
        :param str path_to_csv: Path to Csv-File containing twitter-information
        
        """
        assert on_key == 'RT' or on_key == 'QT' or on_key == 'hashtag', print('Graph not known')

        print('Loading file')
        dataframe = pd.read_csv(path_to_csv, low_memory=False)

        if on_key == 'RT':
            print('Creating Retweet Graph')
            dataframe = dataframe[dataframe.RT == 1]
            self.dataframe = dataframe
            for row in dataframe.itertuples():
                self.add_node(row.r_uid, row.r_screen_name)
                self.add_edge(row.r_uid, row.r_screen_name, row.u_id, row.screen_name)
            self.add_stats(degree=True, betweenness=False, closeness=False)
            self.print_info()

        elif on_key == 'QT':
            print('Creating Quoting Graph')
            dataframe = dataframe[dataframe.QT == 1]
            self.dataframe = dataframe
            for row in dataframe.itertuples():
                self.add_node(row.q_uid, row.q_screen_name)
                self.add_edge(row.q_uid, row.q_screen_name, row.u_id, row.screen_name)
            self.add_stats(degree=True, betweenness=False, closeness=False)
            self.print_info()


        elif on_key == 'hashtag':
            print('Creating Common-Hashtag Graph')

            dataframe = dataframe[dataframe.hashtags != '[]']
            self.dataframe = dataframe

            hashtags_nr = dict()
            iter = 1

            for row in dataframe.itertuples():
                hashtags = eval(row.hashtags)
                if len(hashtags) > 1:
                    for hashtag in hashtags:
                        if str(hashtag) not in hashtags_nr:
                            hashtags_nr[str(hashtag)] = iter
                            nr = iter
                            iter += 1
                        else:
                            nr = hashtags_nr[str(hashtag)]
                        self.add_node(nr, str(hashtag))
                    combos = itertools.combinations(hashtags, 2)
                    for combo in combos:
                        nr_1 = hashtags_nr[combo[0]]
                        nr_2 = hashtags_nr[combo[1]]
                        self.add_edge(nr_1, str(combo[0]), nr_2, str(combo[1]))
            self.add_stats(degree=True, betweenness=False, closeness=False)
            self.print_info()


    def draw_graph(self,
                   output_path,
                   max_size_of_nodes=100,
                   min_size_of_nodes=5,
                   dpi=1000,
                   layout_algorithm='graphviz',
                   display_communities=True,
                   layout_iterations=1000,
                   adjust_sizes=True,
                   linlog_mode=False,
                   draw_edges=True,
                   edge_line_widths=0.2,
                   include_node_labels=True,
                   label_font_size=1,
                   node_line_size=0.2):
        """
        Wrapper function to plot the graph-object represented by this class.
        
        :param str output_path: Path to destination of plot
        
        :param int max_size_of_nodes: Maximum size of a node
        
        :param int min_size_of_nodes: Minimum size of a node
        
        :param int dpi: dots per inch of the output-file
        
        :param str layout_algorithm: Which layout algorithm to use. either 'forceatlas2' or 'graphviz'
        
        :param bool display_communities: Color all members of a community in the same color
        
        :param int layout_iterations: Number of iterations that the layout-algorithm gets applied
        
        :param bool adjust_sizes: Adjust the layout to the size of the nodes == No overlapping
        
        :param bool linlog_mode: Calculate layout with log
        
        :param bool draw_edges: Draw the edges or not
        
        :param float edge_line_widths: thickness of the edges
        
        :param bool include_node_labels: draw the node labels or not
        
        :param float label_font_size: size of label font
        
        :param float node_line_size: thickness of the node borderline
        
        """

        # Make sure no other layout is chosen than the specified
        assert layout_algorithm == 'force' or layout_algorithm == 'graphviz', print("Unknown Layout algorithm")
        assert dpi <= 1000, print("No more than 1000 dpi allowed")

        print("Printing graph to " + output_path)

        # Apply edge_line_widths to all edges
        size_of_nodes = [(max_size_of_nodes + min_size_of_nodes / 2)] * len(self.graph.nodes())
        label_of_nodes = dict()
        edge_line_widths = [edge_line_widths] * len(self.graph.nodes())

        # Calculate smallest and biggest node
        t = 0
        smallest_node_weight = 10
        biggest_node_weight = 10
        for node, data in self.graph.nodes(data=True):
            if data['weight'] < smallest_node_weight:
                smallest_node_weight = data['weight']
            if data['weight'] > biggest_node_weight:
                biggest_node_weight = data['weight']

        # According to actual weight-size of smallest and biggest nodes, calculate relative size in plot according to
        # specified max_size_of_nodes and min_size_of_nodes
        for node, data in self.graph.nodes(data=True):
            size_of_nodes[t] = min_size_of_nodes + (
                ((data['weight'] - smallest_node_weight) * (max_size_of_nodes - min_size_of_nodes)) / (
                    biggest_node_weight - smallest_node_weight))
            label_of_nodes[node] = data['label']
            t += 1

        # Calculate communities if not already done
        if display_communities:
            if not self.community_partitions:
                self._get_communities()
            values = [self.community_partitions.get(node) for node in self.graph.nodes()]
        else:
            values = []

        # Apply layout algorithm
        if layout_algorithm == 'force':
            layout = forceatlas2_networkx_layout(self.graph, niter=layout_iterations, adjustSizes=adjust_sizes,
                                                 linlogMode=linlog_mode,
                                                 sizes=size_of_nodes)
        elif layout_algorithm == 'graphviz':
            try:
                layout = graphviz_layout(self.graph, prog='neato', args='-Goverlap=true')
            except:
                print('Error calculating layout...try with smaller graph')


        # Write titel of plot
        graph_label = '"' + self.name_of_graph + '"' + '\n \n' + self.pruning_information.to_string()

        # Create legend to community color
        legend_dict = OrderedDict()
        for number in values:
            name = 'Community ' + str(number)
            legend_dict[name] = number

        if len(legend_dict) > 20:
            temp = list(islice(legend_dict, 20))
            temp_ordered = OrderedDict()
            for t in temp:
                temp_ordered[t] = legend_dict[t]
            legend_dict = temp_ordered

        paired = plt.get_cmap('Paired')

        if values:
            cNorm = colors.Normalize(vmin=0, vmax=max(values))
            scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=paired)

            f = plt.figure(1)
            ax = f.add_subplot(1, 1, 1)
            for label in legend_dict:
                ax.plot([0], [0], color=scalarMap.to_rgba(legend_dict[label]), label=label)

            plt.legend(loc='center left', prop={'size': 6},bbox_to_anchor=(-0.15, 0.5))

            # If graph-object not empty, draw nodes
            nx.draw_networkx_nodes(self.graph, pos=layout, node_size=size_of_nodes, node_color=values,
                                   linewidths=node_line_size, cmap=paired, ax=ax)

            t = plt.gca()
            t.collections[0].set_edgecolor("#000000")

            plt.axis("off")

            # Draw edges and labels
            if draw_edges:
                nx.draw_networkx_edges(self.graph, pos=layout, arrows=False, width=edge_line_widths)
            if include_node_labels:
                nx.draw_networkx_labels(self.graph, pos=layout, labels=label_of_nodes, font_size=label_font_size,
                                        font_weight='bold')
        # Plot and save the graph
        plt.suptitle(graph_label, size=6)
        plt.savefig(output_path, format="PNG", dpi=dpi)
        print("Writing graph to file successfull")
        plt.close()


"""
The following methods __btwn_pool, __partitions and calculate_betweenness where taken from
https://blog.dominodatalab.com/social-network-analysis-with-networkx/ (21.4.17 - 13:24)

Author: Manojit Nandi, 14.7.15
Modified: Lenz Baumann 21.4.17
"""

def _btwn_pool(G_tuple):
    """
    Wrapper for the method nx.betweenness_centrality_source.
    :param G_tuple: nodes of graph G 
    :return dict:  nodes with their betweenness centrality
    """
    return nx.betweenness_centrality_source(*G_tuple)


def _partitions(nodes, n):
    """
    Splits the nodes up into n different partitions
    :param nodes: 
    :param int n: 
    :return tuple: the partitions
    """
    nodes_iter = iter(nodes)
    while True:
        partition = tuple(itertools.islice(nodes_iter, n))
        if not partition:
            return
        yield partition


def calculate_betweenness(graph):
    """
    Calculates the betweeness of all nodes of the graph in parallel.
    ATTENTION: May take a long time
    
    :param nx.Graph graph: 
    
    :return list of int: Returns the betweenness for each node
     
    """
    p = Pool(processes=None)
    part_generator = 4 * len(p._pool)
    node_partitions = list(_partitions(graph.nodes(), int(len(graph) / part_generator)))
    num_partitions = len(node_partitions)

    bet_map = p.map(_btwn_pool,
                    zip([graph] * num_partitions,
                        [True] * num_partitions,
                        [None] * num_partitions,
                        node_partitions))

    bt_c = bet_map[0]
    for bt in bet_map[1:]:
        for n in bt:
            bt_c[n] += bt[n]
    p.close()
    return bt_c





"""
Unshortens URls of a given Csv-file with the headers 'url' and 'mediaurl'.
Twitter API is documented here: https://dev.twitter.com/overview/api/tweets
It will create a additional logfile in ~/your_directory/input.log
"""

import requests
import logging
import os
import pandas
import numpy as np



# Define Statistic Values
lineNum = 1
deleted_tweet_count = 0
failed_urls = set()
lastid = "None"
urls_dict = dict()


def _unshorten_url(url, tweet_id, logger, timeout=5):
    """
    Takes a url and follows it until redirecting ends or the specified timeout is over.
    The parameters 'tweet_id' and 'logger' are for logging purposes only.
    To save time,every successful unshortening will be stored in a dict for later re-use of shortened-url
    
    :param str url: The url to be unshortened if possible
    
    :param str tweet_id: The tweet_id of tweet containing url
    
    :param logging.Logger logger: The logger that logs all failed urls
    
    :param int timeout: Amount of seconds for url to respond
    
    :return: returns a url as string
    
    """
    if url in urls_dict.keys():
        return urls_dict[url]
    else:
        try:
            answer = requests.head(url, allow_redirects=True, **{'timeout': timeout}).url
            urls_dict[url] = answer
            return answer
        except Exception:
            logger.info(str(tweet_id) + "," + str(url))
            urls_dict[url] = url
            failed_urls.add(url)
            return url


def unshorten_urls(input_path, output_path, timeout = 5):
    """
    Main function of the module 'Unshortening'. Takes path to a Csv-file coming from the module 'Processing' and the
    path for the output-file.
    
    :param str input_path: Path to input Csv-File
    
    :param str output_path: Path to destination of output Csv-File containing the unshortened urls
    
    :param int timeout: Specify the timeout for the __unshorten_url function
        
    """

    #Setting up the logger
    dir_name, _ = os.path.splitext(output_path)
    logger = logging.getLogger(dir_name)
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler("{0}.log".format(dir_name), mode='w')
    fh.setLevel(logging.INFO)
    fh_format = logging.Formatter('%(asctime)s, %(message)s')
    fh.setFormatter(fh_format)
    logger.addHandler(fh)
    fh.setFormatter(logging.Formatter('%(message)s'))
    logger.info("time,tweet_id, failed_url")
    fh.setFormatter(logging.Formatter('%(asctime)s, %(message)s'))

    #Opening the input and the output-paths
    f = open(input_path)

    #Read the Csv-File
    df = pandas.read_csv(f)

    #Iterate over all rows containing urls and try to unshorten them. If successfull replace the old urls with the new
    #unshortened urls.
    for i, row in df.iterrows():
        try:
            n_urls = eval(row.urls)
            urls_temp = [_unshorten_url(a, row['t_id'], logger,timeout=timeout) for a in n_urls if len(n_urls) != 0]
            df.set_value(i, 'urls', repr(urls_temp))

            if row.media != 0:
                n_mediaurls = eval(row.mediaurls) if not pandas.isnull(row.mediaurls) else []
                mediaurls_temp = [_unshorten_url(a, row['t_id'], logger,timeout=timeout) for a in n_mediaurls if len(n_mediaurls) != 0]
                df.set_value(i, 'mediaurls', repr(mediaurls_temp))
            else:
                df.set_value(i, 'mediaurls', np.nan)
        except SyntaxError:
            pass

    #Write unshortened dataframe to Csv-File again.
    o = open(output_path, 'w')
    df.to_csv(o, mode='w', index=False)

    #Write logger summary on amount of failed urls
    fh.setFormatter(logging.Formatter('%(message)s'))
    logger.info("\n")
    logger.info("----------------------------------------------------------------")
    logger.info("\n")
    logger.info("Summary")
    logger.info("Failed urls: " + str(len(failed_urls)))
    logging.shutdown()



"""
Takes a Csv-File and splits it up into a specified number of time-windows. Creates Csv-Files for each window and saves
them into older 'Csv-Slices'.
Contains two methods: slice_dataset and create_overview. Create overview should only be called after slice_dataset has
been called at least once.
"""

import os
import pandas as pd
import pytz
from communitweet.GraphFactory import GraphFactory


def slice_dataset(input_path, number_of_ranges):
    """
    Slices up the Csv-File from the given path into specified number of windows. The windows have an overlap of a third 
    of the timerange each window has
    
    :param str input_path: Path to Csv-File
    
    :param int number_of_ranges: Number of windows and Csv-Files that should be generated
        
    """

    print('Creating Directories')

    # Create all neccessary directories
    try:
        dir_name = os.path.dirname(input_path)
        path_name_csv = os.path.join(dir_name, 'Csv_Slices')
        os.makedirs(path_name_csv, exist_ok=True)
        path_name_graphml = os.path.join(dir_name, 'Slices_Graphml')
        os.makedirs(path_name_graphml, exist_ok=True)
        path_name_png = os.path.join(dir_name, 'Slices_PNG')
        os.makedirs(path_name_png, exist_ok=True)
        path_name_info = os.path.join(dir_name, 'Slices_Info')
        os.makedirs(path_name_info, exist_ok=True)
    except:
        print('Problem Creating Folders...Exiting')
        exit(1)

    print('Loading CSV ... ')

    # Load Csv-File
    df = pd.read_csv(input_path)

    # Create index over time
    df.set_index(pd.DatetimeIndex(df['created_at'], tz=pytz.UTC), inplace=True)

    # Sort index
    df.sort_index(inplace=True, ascending=True)

    # Evaluate length of time slices
    first = df.first_valid_index()
    last = df.last_valid_index()
    date_range = pd.date_range(first, last, freq='1S', tz=pytz.UTC)
    number_of_slices = number_of_ranges
    period = int(round(len(date_range) / number_of_slices, 0))
    third_of_period = int(round(period / 3, 0))
    run = 0
    list_of_ranges = []

    print('Creating ' + str(number_of_ranges) + ' time ranges ...')

    # Evaluate start and end-points for each time-slice
    while run <= len(date_range):
        if run - third_of_period < 0:
            list_of_ranges.append((date_range[run], date_range[run + period]))
        else:
            if not run + period > len(date_range)-1:
                list_of_ranges.append((date_range[run], date_range[run + period]))
            else:
                list_of_ranges.append((date_range[run], date_range[len(date_range) - 1]))
                break
        run += period - third_of_period

    print('Creating CSV-Slices ...')

    # Create time slices from Csv-Input-File
    for (start, end) in list_of_ranges:
        df_temp = df[start:end]
        name_of_slice = str(start) + ' - ' + str(end) + ".csv"
        total_name = os.path.join(path_name_csv, name_of_slice)
        df_temp.reset_index(inplace=True, drop=True)
        df_temp.to_csv(open(total_name, 'w'), index=False)


def create_overview(input_path,node_degree_threshold = 100,edge_weight_minimum = 0, graph_identifier = 'QT',
                    draw_edges = True,
                    ):
    """
    Procedure to produce several overview files out of a given Csv-File containing tweet-data.
    Generates:
    
    - Graph-file according to graph_identifier and saving it in the folder 'Slices_Graphml'
    
    - PNG-File plotting graph to 'Slices_PNG' 
    
    - TXT-File containing information on the graph and its communities and saving it in the folder 'Slices-Info'
    
    
    :param str input_path: Path to directory containing folder of Csv-Files.
    
    :param int node_degree_threshold: Percentage of nodes that should be kept in the graph, after pruning them by the degree of the node
    
    :param int edge_weight_minimum: Minimum weight each node in th graph should have to be kept
    
    :param str graph_identifier: Defines what graph should be drawn on the file. Can be either 'hashtag', 'QT' or 'RT'
    
    :param boolean draw_edges: Should the edges on the graph be drawn or not
    
    """
    print('Creating Graphml, CSV, PNG and Info-Files ...')

    # Get path to CSV-Slices-Folder from input path
    path_name_csv = os.path.join(input_path, 'Csv_Slices')
    for filename in os.listdir(path_name_csv):
        if filename.endswith('csv'):

            # Get filename of Csv-File
            total_file_dest = os.path.join(path_name_csv, filename)

            # Instantiate GraphFactory
            gf = GraphFactory(str(filename))

            # Create a graph according to graph_identifier
            gf.create_graphs_from_dataframe(graph_identifier, total_file_dest)

            # Change directory
            total_file_dest = os.path.abspath(os.path.join(path_name_csv, os.pardir))
            total_file_dest = os.path.join(total_file_dest, 'Slices_Graphml')
            temp = os.path.join(total_file_dest, filename.rstrip('.csv') + '.graphml')

            # Prune graph according to specified parameters
            gf.prune_graph(edge_weight=edge_weight_minimum, degree_thresh=0, weight_thresh=0, remove_isolates=False)
            threshold = gf.get_threshold(node_degree_threshold, 2, 'degree_thresh')
            gf.prune_graph(edge_weight=0, degree_thresh=threshold, weight_thresh=0, remove_isolates=True)

            # Write graph to graphml-file
            gf.write_graphml(temp)

            # Change directory
            total_file_dest = os.path.abspath(os.path.join(path_name_csv, os.pardir))
            total_file_dest = os.path.join(total_file_dest, 'Slices_PNG')
            temp = os.path.join(total_file_dest, filename.rstrip('.csv') + '.png')

            gf.print_info()

            # Draw Graph according to specified parameters
            gf.draw_graph(temp, 100, 5, layout_algorithm='graphviz', draw_edges=draw_edges)

            # Change directory
            total_file_dest = os.path.abspath(os.path.join(path_name_csv, os.pardir))
            total_file_dest = os.path.join(total_file_dest, 'Slices_Info')
            temp = os.path.join(total_file_dest, filename.rstrip('.csv') + '.txt')

            # Write out graph and community information
            gf.write_out_community_user_info(temp)



from setuptools import setup

setup(name='communitweet',
      version='1.2',
      description='A package that wraps tools for harvesting, hydrating, cleaning and visualize twitter data',
      url='https://github.com/HolzmanoLagrene/FacharbeitLenzBaumannHS16FS17',
      author='Lenz Baumann',
      author_email='lnzbmnn@gmail.com',
      license='MIT',
      packages=['communitweet'],
      install_requires=['bs4', 'pandas', 'numpy', 'python-louvain', 'matplotlib', 'networkx', 'twarc'],
      include_package_data=True,
      zip_safe=False)




"""
Adds Information on demographic data (Population, Race and Poverty) to database
"""

from urllib import request
from bs4 import BeautifulSoup
import json
import pandas as pd
import os


def _get_population(dataframe, longitude, latitude):
    """
    This function gets a dataframe and geo-coordinates and gets the FIPS-Codes and the Names of the state and the county
    the coordinates are placed in. On the basis of this information the exact population for the county in the year 2010
    can be determined.
    
    :param pd.Dataframe dataframe: Dataframe containing population information for each county
    
    :param float longitude: Longitude coordinate of tweet
    
    :param float latitude: Latitude coordinate of tweet
    
    :return dict of (str,str): In case of successfull identification of county, a dict containing names and FIPS-Codes 
    as well as population information
    
    """

    # Call to data.fcc.gov/api is made to determine county and state
    url1 = 'http://data.fcc.gov/api/block/find?latitude={0}&longitude={1}&showall=false&format=json'.format(
        str(latitude),
        str(longitude))
    with request.urlopen(url1) as response:
        try:
            html = response.read()
            b4 = BeautifulSoup(html, 'lxml')
            a = json.loads(str(b4.get_text()))

            # Api-Information is evaluated
            if a['County']['FIPS'] is not None:
                state_FIPS = a['State']['FIPS']
                county_FIPS = a['Block']['FIPS'][2:5]
                combined_FIPS = str(state_FIPS) + str(county_FIPS).zfill(3)

                # Population data gets determined
                df = dataframe[dataframe.Combined_FIPS == int(combined_FIPS)]
                pop = df['Population'].values[0]
                name = df['Name'].values[0]
                county_name, state_name = name.split(', ')
                return {'State_Name': state_name, 'County_Name': county_name, 'Population': pop,
                        'State_FIPS': state_FIPS,
                        'County_FIPS': county_FIPS}
            else:
                print('Coordinates not in the US')
                return None
        except:
            return None


def _get_racial_statistics(dataframe, state_code, county_code, population):
    """
    Calculates percentages of Hispanic, Black and White people according county using the total amount of people in the 
    county
    
    :param pd.Dataframe dataframe: Dataframe containing racial information on each county
    
    :param int state_code:
    
    :param int county_code: 
    
    :param int population: 
    
    :return dict of (str,float): 
    
    """

    # Dataframe gets narrowed down to relevant information
    df = dataframe[dataframe.STATE == int(state_code)]
    df = df[df.COUNTY == int(county_code)]

    # Percentage of hispanic people
    df_hisp = df[df.ORIGIN == 2]
    perc_hisp = float(round(pd.DataFrame(df_hisp.groupby([df.IMPRACE])['RESPOP'].sum())['RESPOP'].sum() / population * 100, 2))

    # Percentage of black people
    df_black = df[df.IMPRACE == 2]
    perc_black = float(round(pd.DataFrame(df_black.groupby([df.ORIGIN])['RESPOP'].sum())['RESPOP'].sum() / population * 100,
                       2))

    # Percentage of white people
    df_white = df[df.IMPRACE == 1]
    perc_white = float(round(pd.DataFrame(df_white.groupby([df.ORIGIN])['RESPOP'].sum())['RESPOP'].sum() / population * 100,
                       2))

    return {'hispanic': perc_hisp, 'black': perc_black, 'white': perc_white}


def _get_poverty_statistics(dataframe, state_code, county_code):
    """
    Calculates poverty-rate of people living below certain threshold
    
    :param dataframe: Containing poverty data
    
    :param int state_code: 
    
    :param int county_code: 
    
    :return float:
    
    """
    dataframe.columns = ['STATE', 'COUNTY', 'NAME', 'POV']
    df = dataframe[dataframe.STATE == int(state_code)]
    if len(df) > 0:
        df = df[df.COUNTY == int(county_code)]
        return float(df['POV'].values[0])
    else:
        return 0


def add_demographic_statistics(input_path, output_path):
    """
    Takes path to Csv-File after beeing processed by module 'Processing' and adds demographic information on race,
    population and poverty to it according to geo-information - if any.
    
    :param str input_path: Path to Csv-Input-File
    
    :param str output_path: Path to destination of output-file. File gets overwritten if already existing and generated if not
    
    """

    # Open input file
    df_twitter_data = pd.read_csv(open(input_path))

    # Add column headers to input Csv-file
    df_twitter_data['has_stats'] = ''
    df_twitter_data['State_FIPS'] = ''
    df_twitter_data['County_FIPS'] = ''
    df_twitter_data['State_Name'] = ''
    df_twitter_data['County_Name'] = ''
    df_twitter_data['Population'] = ''
    df_twitter_data['Perc_Hisp'] = ''
    df_twitter_data['Perc_Black'] = ''
    df_twitter_data['Perc_White'] = ''
    df_twitter_data['Perc_Poverty'] = ''

    # Open datasheets containing population, racial and poverty data
    this_dir,_ = os.path.split(__file__)

    racial_data_dir = os.path.join(this_dir,'data_sheets','race_data.csv')
    df_racial_data = pd.read_csv(open(racial_data_dir))

    poverty_data_dir = os.path.join(this_dir,'data_sheets','poverty_data.csv')
    df_poverty_data = pd.read_csv(open(poverty_data_dir),
                                  usecols=['State FIPS', 'County FIPS', 'Name', 'Poverty Percent All Ages'])

    poverty_data_dir = os.path.join(this_dir,'data_sheets','population_data.csv')
    df_population_data = pd.read_csv(open(poverty_data_dir), delimiter=';')

    # Iterate through rows of input-file
    for i, row in df_twitter_data.iterrows():

        # If tweet contains geo-information, evaluate longitude and latitude and update it in the file if necessary
        if row.is_geo == 1:
            geo_cord_dict = _get_population(df_population_data, int(row.geo_long), int(row.geo_lat))

            if geo_cord_dict is not None:
                # Get racial and poverty data
                RaceDict = _get_racial_statistics(df_racial_data, geo_cord_dict['State_FIPS'], geo_cord_dict['County_FIPS'],
                                                  geo_cord_dict['Population'])
                PovertyPerc = _get_poverty_statistics(df_poverty_data, geo_cord_dict['State_FIPS'],
                                                      geo_cord_dict['County_FIPS'])

                # Update file with new information
                df_twitter_data.set_value(i, 'has_stats', 1)
                df_twitter_data.set_value(i, 'County_Name', geo_cord_dict['County_Name'])
                df_twitter_data.set_value(i, 'County_FIPS', geo_cord_dict['County_FIPS'])
                df_twitter_data.set_value(i, 'State_Name', geo_cord_dict['State_Name'])
                df_twitter_data.set_value(i, 'State_FIPS', geo_cord_dict['State_FIPS'])
                df_twitter_data.set_value(i, 'Population', repr(geo_cord_dict['Population']))
                df_twitter_data.set_value(i, 'Perc_Hisp', repr(RaceDict['hispanic']))
                df_twitter_data.set_value(i, 'Perc_Black', repr(RaceDict['black']))
                df_twitter_data.set_value(i, 'Perc_White', repr(RaceDict['white']))
                df_twitter_data.set_value(i, 'Perc_Poverty', repr(PovertyPerc))


        # Write new data to output destination
    df_twitter_data.to_csv(open(output_path, 'w'), mode='w', index=False)
